;;;-*- Mode: Lisp; Package: OCI -*-

;; OCI module
;; Version 0.5 (experimental)
;; Copyright (C) Paul Meurer 2001. All rights reserved.
;; paul.meurer@hit.uib.no
;;
;; Documentation and the license agreement can be found in file 
;; "sql-odbc-documentation.lisp".
;; Bug reports and suggestions are highly welcome.

;; See documentation at:
;; http://www.oradoc.com/ora817/appdev.817/a76975/oci15rel.htm

(in-package :oci)

(defparameter +max-precision+ 
  #+mcl 512
  #-mcl 4001)

(defun handle-error (errhp)
  (%with-temporary-allocation ((error-message :string 512)
                               (error-code :signed-long))
    (OCIErrorGet errhp 1 (%null-ptr) error-code error-message 512 ;; msg-length
                 +OCI-HTYPE-ERROR+)
    (values
     (%get-cstring error-message)
     (%get-long error-code))))

(defmacro with-error-handling ((errorhp &rest ignore-error-codes) oci-call &body body)
  (let ((result-code (gensym))
        (error-code (gensym))
        (error-message (gensym)))
    `(let ((,result-code ,oci-call))
       ;; *** is this a bug in allegro or in my code??
       (case ,result-code
         (#.+OCI-SUCCESS+
          (progn ,result-code ,@body))
         (#.+OCI-SUCCESS-WITH-INFO+
          #+not-yet
          (when ,print-info
            (multiple-value-bind (error-message sql-state)
                (handle-error (or ,henv (%null-ptr))
                              (or ,hdbc (%null-ptr))
                              (or ,hstmt (%null-ptr)))
              (warn "[ODBC info] ~a state: ~a"
                    ,result-code error-message
                    sql-state)))
          (print "error: +SQL-SUCCESS-WITH-INFO+")
          (progn ,result-code ,@body))
         (#.+OCI-NEED-DATA+
          (error "[OCI error] Need data. *prelim*"))
         (#.+OCI-INVALID-HANDLE+
          (error "[OCI error] Invalid handle"))
         (#.+OCI-STILL-EXECUTING+
	  (if (find :signal-still-executing ',ignore-error-codes)
	      :still-executing
	    (error "[OCI error] Still executing")))
         (#.+OCI-CONTINUE+
          (error "[OCI error] Continue"))
         (#.+OCI-ERROR+
          (multiple-value-bind (,error-message ,error-code) (handle-error ,errorhp)
            (if (or (eq (car ',ignore-error-codes) t)
                    (find ,error-code ',ignore-error-codes))
                ,error-code
              (error "[OCI error] ~a" ,error-message))))
         (otherwise
          (progn ,result-code ,@body))))))

#-oracle-i
(defun %new-environment-handle (oci-err)
  (%with-sql-pointer (oci-env)
    (with-error-handling (oci-err)
        (OCIInitialize  +OCI-THREADED+ (%null-ptr) (%null-ptr) (%null-ptr) (%null-ptr)))
    (with-error-handling (oci-err)
        (OCIEnvInit oci-env +OCI-DEFAULT+ 0 (%null-ptr))
      (%get-ptr oci-env))))

#+oracle-i
(defun %new-environment-handle (oci-err)
  (%with-sql-pointer (oci-env)
    (with-error-handling (oci-err)
        (OCIEnvCreate oci-env +OCI-THREADED+
                      (%null-ptr) (%null-ptr) (%null-ptr) (%null-ptr)
                      0 (%null-ptr))
      (%get-ptr oci-env))))

(defun %new-error-handle (oci-env)
  (%with-sql-pointer (oci-err)
    (with-error-handling ((%null-ptr))
        (OCIHandleAlloc oci-env oci-err +OCI-HTYPE-ERROR+ 0 (%null-ptr))
      (%get-ptr oci-err))))

(defun %new-server-handle (oci-env oci-err)
  (%with-sql-pointer (oci-server)
    (with-error-handling (oci-err)
        (OCIHandleAlloc oci-env oci-server +OCI-HTYPE-SERVER+ 0 (%null-ptr))
      (%get-ptr oci-server))))

(defun %new-service-context-handle (oci-env oci-err)
  (%with-sql-pointer (oci-svcctx)
    (with-error-handling (oci-err)
        (OCIHandleAlloc oci-env oci-svcctx +OCI-HTYPE-SVCCTX+ 0 (%null-ptr))
      (%get-ptr oci-svcctx))))

(defun %new-session-handle (oci-env oci-err)
  (%with-sql-pointer (oci-session)
    (with-error-handling (oci-err)
        (OCIHandleAlloc oci-env oci-session +OCI-HTYPE-SESSION+ 0 (%null-ptr))
      (%get-ptr oci-session))))

(defun %oci-server-attach (oci-server oci-err server-name)
  (let* ((len (length server-name)))
    (with-cstr (server-ptr server-name)
      (with-error-handling (oci-err)
          (OCIServerAttach oci-server oci-err server-ptr len 0)))))

(defun %oci-server-detach (oci-server oci-err)
  (with-error-handling (oci-err)
      (OCIServerDetach oci-server oci-err +OCI-DEFAULT+)))

(defun %oci-attr-set (handle handle-type attribute-pointer size attribute-type oci-err)
  (with-error-handling (oci-err)
      (OCIAttrSet handle handle-type attribute-pointer size attribute-type oci-err)))

#+lispworks
(defun %oci-attr-set-string (handle handle-type attribute-pointer size attribute-type oci-err)
  (%oci-attr-set handle handle-type attribute-pointer size attribute-type oci-err))

#+allegro
(defun %oci-attr-set-string (handle handle-type attribute-pointer size attribute-type oci-err)
  (with-error-handling (oci-err)
      (OCIAttrSet-string handle handle-type attribute-pointer size attribute-type oci-err)))

(defun %oci-attr-get (handle handle-type attribute-pointer size-ptr attribute-type oci-err)
  (with-error-handling (oci-err)
      (OCIAttrGet handle handle-type attribute-pointer size-ptr attribute-type oci-err)))

#+lispworks
(defun %oci-attr-get-string (handle handle-type attribute-pointer size-ptr attribute-type oci-err)
  (%oci-attr-get handle handle-type attribute-pointer size-ptr attribute-type oci-err))

#+allegro
(defun %oci-attr-get-string (handle handle-type attribute-pointer size-ptr attribute-type oci-err)
  (with-error-handling (oci-err)
      (OCIAttrGet-string handle handle-type attribute-pointer size-ptr attribute-type oci-err)))

(defun %oci-session-begin (oci-server oci-err oci-auth credits mode)
  (with-error-handling (oci-err)
      (OCISessionBegin oci-server oci-err oci-auth credits mode)))

(defun %oci-session-end (oci-service-context oci-error)
  (with-error-handling (oci-error)
      (OCISessionEnd oci-service-context oci-error (%null-ptr) +OCI-DEFAULT+)))

(defun %oci-connect (oci-env oci-server oci-service-context oci-err server-name user-id password)
  (%oci-server-attach oci-server oci-err server-name)
  (%oci-attr-set oci-service-context +OCI-HTYPE-SVCCTX+ oci-server 0 +OCI-ATTR-SERVER+ oci-err)
  (let ((oci-session (%new-session-handle oci-env oci-err)))
    (with-cstr (user-ptr user-id)
      (%oci-attr-set-string oci-session +OCI-HTYPE-SESSION+
			    user-ptr (length user-id)
                     +OCI-ATTR-USERNAME+ oci-err))
    (with-cstr (password-ptr password)
      (%oci-attr-set-string oci-session +OCI-HTYPE-SESSION+
			    password-ptr (length password)
                     +OCI-ATTR-PASSWORD+ oci-err))
    (%oci-session-begin oci-service-context oci-err oci-session +OCI-CRED-RDBMS+ +OCI-DEFAULT+)
    (%oci-attr-set oci-service-context +OCI-HTYPE-SVCCTX+ oci-session 0
                   +OCI-ATTR-SESSION+ oci-err)
    (%with-temporary-allocation ((blocking-mode-ptr :short)
				 #+ignore
				 (blocking-mode-ptr #+allegro :ptr
						    #+lispworks(:pointer :char))
				 #+ignore
				 (bm-ptr-length :unsigned-long))
	(%put-word blocking-mode-ptr 1)
	#-ignore
	(%oci-attr-set oci-service-context +OCI-HTYPE-SVCCTX+ blocking-mode-ptr
		       0 +OCI-ATTR-NONBLOCKING-MODE+ oci-err)
        (%oci-attr-get oci-service-context +OCI-HTYPE-SVCCTX+ blocking-mode-ptr
		       (%null-ptr) +OCI-ATTR-NONBLOCKING-MODE+ oci-err)
	#+debug
	(format t "~%blocking-mode: ~d"
		#+ignore
		(%get-string (%get-ptr blocking-mode-ptr)
				      :end (%get-unsigned-long bm-ptr-length))
		(%get-short blocking-mode-ptr)))
    oci-session))

(defun %oci-disconnect (oci-server oci-service-context oci-error)
  ;; this gives invalid handle error
  ;;(%oci-session-end oci-service-context oci-error)
  (%oci-server-detach oci-server oci-error)
  (%free-oci-handle oci-server oci-error +OCI-HTYPE-SERVER+)
  (%free-oci-handle oci-service-context oci-error +OCI-HTYPE-SVCCTX+)
  (%free-oci-handle oci-error oci-error +OCI-HTYPE-ERROR+))

(defun %free-oci-handle (oci-handle oci-err type)
  (with-error-handling (oci-err)
      (OCIHandleFree oci-handle type)))

(defun %free-statement (oci-stmt oci-err)
  (%free-oci-handle oci-stmt oci-err +OCI-HTYPE-STMT+))

(defun %oci-trans-start (oci-service-context oci-error timeout)
  (OCITransStart oci-service-context oci-error timeout +OCI-TRANS-NEW+))

(defun %oci-trans-commit (oci-service-context oci-error)
  (OCITransCommit oci-service-context oci-error +OCI-DEFAULT+))

(defun %oci-trans-rollback (oci-service-context oci-error)
  (OCITransRollback oci-service-context oci-error +OCI-DEFAULT+))

; col-nr is zero-based in Lisp
(defun %oci-define-by-pos (oci-stmt oci-def-ptr oci-err pos data-ptr size oci-type 
                                    out-len-ptr return-code-ptr)
  (with-error-handling (oci-err)
      (OCIDefineByPos oci-stmt oci-def-ptr oci-err pos data-ptr size oci-type (%null-ptr) 
                      out-len-ptr return-code-ptr +OCI-DEFAULT+)))

(defun %bind-column (oci-stmt oci-def oci-err column-nr oci-type data-ptr
                              size out-len-ptr return-code-ptr)
  (%oci-define-by-pos oci-stmt oci-def oci-err 
                      (if (eq column-nr :bookmark) 0 (1+ column-nr))
                      data-ptr size oci-type out-len-ptr return-code-ptr))

(defun %oci-define-array-of-struct (oci-def oci-err pv-skip ind-skip rl-skip rc-skip)
  (with-error-handling (oci-err)
      (OCIDefineArrayOfStruct oci-def oci-err pv-skip ind-skip rl-skip rc-skip)))

(defun %oci-stmt-fetch (oci-stmt oci-err nrows wait-function)
  (let ((result nil))
    (loop while (eq :still-executing
		    (setf result
			  (with-error-handling
			      (oci-err 1405 :signal-still-executing)
			      (OCIStmtFetch oci-stmt oci-err nrows
					    +OCI-FETCH-NEXT+ +OCI-DEFAULT+))))
	  when wait-function
          do (funcall wait-function))
    result))
  
(defun %new-statement-handle (oci-env oci-err)
  (%with-sql-pointer (oci-stmt)
    (with-error-handling (oci-err)
        (OCIHandleAlloc oci-env oci-stmt +OCI-HTYPE-STMT+ 0 (%null-ptr))
      (%get-ptr oci-stmt))))

(defun %oci-stmt-execute (oci-svc oci-stmt oci-err iterations wait-function)
  (loop while (eq :still-executing
		  (with-error-handling
		   (oci-err :signal-still-executing)
		   (OCIStmtExecute oci-svc oci-stmt oci-err
				   iterations 0 (%null-ptr) (%null-ptr) +OCI-DEFAULT+)))
	when wait-function
        do (funcall wait-function)))

;; returns nil if column does not exist
(defun %oci-param-get (oci-stmt oci-type oci-err oci-param pos)
  (not (eql (with-error-handling (oci-err 24334)
                (OCIParamGet oci-stmt oci-type oci-err oci-param pos))
            24334)))

(defun %describe-column (oci-stmt oci-err column-nr)
  (%with-temporary-allocation ((oci-param-ptr #+allegro :ptr
					      #+lispworks :pointer)
                               (column-name-ptr #+allegro :ptr
						#+lispworks-ignore string-ptr
						#+lispworks(:pointer :char))
                               (column-name-length :unsigned-long)
                               ;;(column-charset-id :unsigned-short) ;; new
                               (column-sql-type :unsigned-short)
                               (column-data-size :unsigned-short) ;; not ub4 !?!
                               (column-precision :unsigned-short)
                               (column-scale :unsigned-short)
                               (column-nullable-p :unsigned-short))
    (when (%oci-param-get oci-stmt +OCI-HTYPE-STMT+ oci-err oci-param-ptr column-nr)
      (let ((oci-param (%get-ptr oci-param-ptr)))
        ;; the sql type of the column/argument
        (%oci-attr-get oci-param +OCI-DTYPE-PARAM+ column-sql-type (%null-ptr)
                       +OCI-ATTR-DATA-TYPE+ oci-err)
        ;; the name of the column/argument
        (%oci-attr-get-string oci-param +OCI-DTYPE-PARAM+
			      column-name-ptr column-name-length
			      +OCI-ATTR-NAME+ oci-err)
        ;; precision if number type
        (%oci-attr-get oci-param +OCI-DTYPE-PARAM+ column-precision (%null-ptr)
                       +OCI-ATTR-PRECISION+ oci-err)
        ;; data size
        (%oci-attr-get oci-param +OCI-DTYPE-PARAM+ column-data-size (%null-ptr)
                       +OCI-ATTR-DATA-SIZE+ oci-err)
        ;; scale if number type
        (%oci-attr-get oci-param +OCI-DTYPE-PARAM+ column-scale (%null-ptr)
                       +OCI-ATTR-SCALE+ oci-err)
        ;; is it null?
        (%oci-attr-get oci-param +OCI-DTYPE-PARAM+ column-nullable-p (%null-ptr)
                       +OCI-ATTR-IS-NULL+ oci-err)
        (values
         t 
         ;; %get-functions are mostly of wrong type. This doesn't matter in LW,
         ;; but will be buggy in other Lisps
         (%get-string (%get-ptr column-name-ptr)
		      :end (%get-unsigned-long column-name-length))
         (%get-signed-word column-sql-type)
         (%get-short column-precision)
         (%get-short column-data-size)
         (%get-signed-word column-scale)
         (%get-signed-word column-nullable-p))))))

(defun %get-column-data-size (oci-stmt oci-err column-nr)
  (%with-temporary-allocation ((oci-param-ptr #+allegro :ptr
					      #+lispworks :pointer)
                               (column-data-size :unsigned-short))
    (when (%oci-param-get oci-stmt +OCI-HTYPE-STMT+ oci-err oci-param-ptr column-nr)
      (let ((oci-param (%get-ptr oci-param-ptr)))
        (%oci-attr-get oci-param +OCI-DTYPE-PARAM+ column-data-size (%null-ptr)
                       +OCI-ATTR-DATA-SIZE+ oci-err)
	(%get-short column-data-size)))))

#|
Oracle stores values of the NUMBER datatype in a variable-length format. The first byte is the exponent and is followed by 1 to 20 mantissa bytes. The high-order bit of the exponent byte is the sign bit; it is set for positive numbers. The lower 7 bits represent the exponent, which is a base-100 digit with an offset of 65. 

Each mantissa byte is a base-100 digit, in the range 1..100. For positive numbers, the digit has 1 added to it. So, the mantissa digit for the value 5 is 6. For negative numbers, instead of adding 1, the digit is subtracted from 101. So, the mantissa digit for the number -5 is 96 (101-5). Negative numbers have a byte containing 102 appended to the data bytes. However, negative numbers that have 20 mantissa bytes do not have the trailing 102 byte. Because the mantissa digits are stored in base 100, each byte can represent 2 decimal digits. The mantissa is normalized; leading zeroes are not stored. 
|#

(defun %get-sqlt-varnum (data-ptr &optional (offset 0))
  (let* ((size (%get-nth-byte data-ptr offset))
         (exp-byte (%get-nth-byte data-ptr (1+ offset)))
         (negative-p (zerop (ash exp-byte -7)))
	 (exponent (- (logand exp-byte 127) 63 (if negative-p (- 2 size) size)))
         (number 0))
    (loop for i from 1 to (1- size)
          for byte = (%get-nth-byte data-ptr (+ 1 i offset))
          do
          (unless (= byte 102)
	    (let ((n (if negative-p (- 101 byte) (1- byte))))
	      (setf number (+ (* number 100) n)))))
    (if negative-p (setf number (- number)))
    ;;(print (list number exp size exponent negative-p))
    (cond ((zerop number)
	   0)
	  ((zerop exponent)
	   number)
	  ((minusp exponent)
	   (float (* number (expt 100 exponent))))
	  (t
	   (* number (expt 100 exponent))))))

#|
The century and year bytes are in an excess-100 notation. Dates Before Common Era (BCE) are less than 100. The era begins on 01-JAN-4712 BCE, which is Julian day 1. For this date, the century byte is 53, and the year byte is 88. The hour, minute, and second bytes are in excess-1 notation. The hour byte ranges from 1 to 24, the minute and second bytes from 1 to 60. If no time was specified when the date was created, the time defaults to midnight (1, 1, 1). 

byte meaning
------------
 1   Century 
 2   Year 
 3   Month 
 4   Day 
 5   Hour 
 6   Minute 
 7   Second
 
|#

(defun %get-sqlt-date (data-ptr &optional (offset 0))
  (let ((century (%get-nth-byte data-ptr (+ 0 offset)))
        (year (%get-nth-byte data-ptr (+ 1 offset)))
        (month (%get-nth-byte data-ptr (+ 2 offset)))
        (day (%get-nth-byte data-ptr (+ 3 offset)))
        (hour (%get-nth-byte data-ptr (+ 4 offset)))
        (minute (%get-nth-byte data-ptr (+ 5 offset)))
        (second (%get-nth-byte data-ptr (+ 6 offset))))
    (encode-universal-time (1- second)
                           (1- minute)
                           (1- hour)
                           day
                           month
                           (+ (* 100 (- century 100))
                              (- year 100)))))

(defun read-data (data-ptr sql-type out-len-ptr return-code-ptr data-size nrows)
  (if (= nrows 1)
      (let ((length (%get-short out-len-ptr))
            (return-code (%get-short return-code-ptr)))
        (cond ((= return-code 1405)
               *null*)
              (t
               (ecase sql-type
                 (#.+SQLT-CHR+ (%get-string data-ptr :end length))
                 (#.+SQLT-AFC+ (%get-string data-ptr :end length))
                 (#.+SQLT-VNU+ (%get-sqlt-varnum data-ptr))
                 ;;(#.+SQLT-NUM+ (%get-sqlt-num data-ptr))
                 (#.+SQLT-INT+ (%get-ptr data-ptr))
                 (#.+SQLT-DAT+ (%get-sqlt-date data-ptr))))))
    (collecting
     (block loop
       (loop for i from 0 to (1- nrows)
	     do
	     (let ((length
		    #+lispworks
		    (fli::foreign-aref out-len-ptr i)
		    #+allegro
		    (ff:fslot-value out-len-ptr i))
		   (return-code
		    #+lispworks
		    (fli::foreign-aref return-code-ptr i)
		    #+allegro
		    (ff:fslot-value return-code-ptr i)))
	       (case return-code
		 (1405
		  (collect *null*))
		 (#.+OCI-NO-DATA+
		  (return-from loop))
		 (otherwise
		  #+ignore
		  (print (list sql-type return-code i length
			       :start (* i (1+ data-size))
			       :end (+ (* i (1+ data-size)) length)))
		  (collect
		   (ecase sql-type
		     (#.+SQLT-CHR+ (%get-string data-ptr
						:start (* i (1+ data-size))
						:end (+ (* i (1+ data-size)) length)))
		     (#.+SQLT-AFC+ (%get-string data-ptr
						:start (* i (1+ data-size))
						:end (+ (* i (1+ data-size)) length)))
		     (#.+SQLT-VNU+ (%get-sqlt-varnum data-ptr (* i 23)))
		     (#.+SQLT-INT+ (%get-ptr data-ptr))
		     (#.+SQLT-DAT+ (%get-sqlt-date data-ptr (* i 8)))))))))))))

(defvar *break-on-unknown-data-type* t)

#|
If you specify the datatype code 2 in the dty parameter of an OCIDefineByPos() call, your program receives numeric data in this Oracle internal format. The output variable should be a 21-byte array to accommodate the largest possible number. Note that only the bytes that represent the number are returned. There is no blank padding or null termination. If you need to know the number of bytes returned, use the VARNUM external datatype instead of NUMBER. See the description of VARNUM for examples of the Oracle internal number format. 
|#

;; SQLT_CHR, SQL_AFC, SQLT_DAT, SQLT_INT, SQLT_UIN, SQLT_FLT, SQLT_PDN,
;; SQLT_BIN, or SQLT_NUM. 


(defun %allocate-bindings (sql-type precision size row-fetch-count)
  (when (> (* row-fetch-count (1+ size)) 32767)
    (error "~%array size ~d bigger than 32767. Reduce row-fetch-count (now ~d)."
           (* row-fetch-count (1+ size)) row-fetch-count))
  (if (= row-fetch-count 1)
      (let ((data-ptr
             (case sql-type
               (#.+SQLT-NUM+ (%new-cstring (1+ size)))
               (#.+SQLT-VNU+ (%new-cstring (1+ size)))
               (#.+SQLT-INT+ (%new-ptr :long))
               (#.+SQLT-DAT+ (%new-cstring 7))
               (#.+SQLT-CHR+ (%new-cstring (1+ size)))
               (#.+SQLT-AFC+ (%new-cstring (1+ size))) ;; ??
               (t
                (when *break-on-unknown-data-type*
                  (break "Unknown data type: SQL type is ~A, precision ~D, size ~D" 
                         sql-type precision size))
                (%new-ptr :ptr (1+ size)))))
            (out-len-ptr (%new-ptr :unsigned-short))
            (def-ptr (%new-ptr :unsigned-short))
            (return-code-ptr (%new-ptr :unsigned-short)))
        (values data-ptr def-ptr out-len-ptr return-code-ptr size #+ignore long-p))
    (let ((data-ptr
           (case sql-type
             (#.+SQLT-NUM+ (%new-cstring (* row-fetch-count (1+ size))))
             (#.+SQLT-VNU+ (%new-cstring (* row-fetch-count (1+ size))))
             (#.+SQLT-INT+ (%new-array :long row-fetch-count))
             (#.+SQLT-DAT+ (%new-cstring (* row-fetch-count (1+ size))))
             (#.+SQLT-CHR+ (%new-cstring (* row-fetch-count size)))
             (#.+SQLT-AFC+ (%new-cstring (* row-fetch-count size)))
             (t 
              (break "Unknown data type: SQL type is ~A, precision ~D, size ~D" 
                     sql-type precision size))))
          (out-len-ptr (%new-array :signed-short row-fetch-count))
          (def-ptr (%new-array :unsigned-short row-fetch-count))
          (return-code-ptr (%new-array :signed-short row-fetch-count)))
      ;;(format t "~%allocated: ~a, ~a" data-ptr return-code-ptr)
      (values data-ptr def-ptr out-len-ptr return-code-ptr size #+ignore long-p))))

(defun %oci-stmt-prepare (oci-stmt oci-err sql)
  (with-cstr (sql-ptr sql)
    (with-error-handling (oci-err)
        (OCIStmtPrepare oci-stmt oci-err sql-ptr (length sql) +OCI-NTV-SYNTAX+ +OCI-DEFAULT+))))

;; EOF